/* Include the controller definition */
#include "quadcopter.h"
/* Function definitions for XML parsing */
#include <argos3/core/utility/configuration/argos_configuration.h>
/* Function definitions for logging */
#include <argos3/core/utility/logging/argos_log.h>

#include <loop_functions/quadcopter_loop_functions/quadcopter_loop_functions.h>

/****************************************/
/****************************************/

/* Altitude to circle to move along the circle */
Real ALTITUDE;

/* X-Axis to circle to move along the circle */
Real XAxisCircle; 
/* 
static const Real XAxisCircle = 5.0f;

/* Y-Axis to circle to move along the circle */
Real YAxisCircle;

/* Radius of the circle */
Real CIRCLE_RADIUS;

/* How many points the robot traverses to move along the X AXIS */
UInt32 CIRCLE_WAYPOINTS; 

/* How many points the robot traverses to move along the Y AXIS */
UInt32 Y_WAYPOINTS ; 

/* How many points the robot traverses to move along the X AXIS */

Real m_NOXWAYPOINT_SLICE;
Real m_XWAYPOINTSLICE;

/* How many points the robot traverses to move along the X AXIS */
static const Real Y_WAYPOINTSLICE= 0.1f; 

/* The angle between each waypoint */
static const CRadians CIRCLE_SLICE(2.0f * ARGOS_PI / CIRCLE_WAYPOINTS);

/* Tolerance for the distance to a target point to decide to do something else */
static const Real PROXIMITY_TOLERANCE = 0.01f;

/****************************************/
/****************************************/


/****************************************/
/****************************************/
CEyeBotCircle::SFoodData::SFoodData() :
   HasFoodItem(false),
   rechargestates(true),
   FoodItemIdx(0),
   nEnergy(0),
   TotalFoodItems(0) {}

void CEyeBotCircle::SFoodData::Reset() {
   HasFoodItem = false;
   rechargestates = true;
   FoodItemIdx = 0;
   nEnergy = 0;
   TotalFoodItems = 0;
}
/****************************************/
/****************************************/

CEyeBotCircle::CEyeBotCircle() :
   m_pcPosAct(NULL),
   m_pcPosSens(NULL),
   m_pcLEDs(NULL),
   m_pcRNG(NULL),
   m_workingstartx(0.0f, 100.0f),
   m_XWAYPOINTSLICE(0.1f),
   m_NOXWAYPOINT_SLICE(0.1f),
   m_CIRCLE_WAYPOINTSX(0.0f),
   m_CIRCLE_WAYPOINTSY(0.0f),
   m_rdmlocx(0.0f),
   m_rdmlocy(0.0f),
   m_sensx(0.0f),
   m_sensy(0.0f),
   m_funWaypoint (0.1f),
   m_fYunWaypoint (0.1f),
   m_nEnergy (0),
   m_pcLightSensor(NULL),
   
   /*
    m_pcRotorSpeed(NULL),
    */
   m_pcRABSens(NULL) {}

/****************************************/
/****************************************/

void CEyeBotCircle::Init(TConfigurationNode& t_node) {
   /*
    * Get sensor/actuator handles
    *
    * The passed string (ex. "quadrotor_position") corresponds to the
    * XML tag of the device whose handle we want to have. For a list of
    * allowed values, type at the command prompt:
    *
    * $ argos3 -q actuators
    *
    * to have a list of all the possible actuators, or
    *
    * $ argos3 -q sensors
    *
    * to have a list of all the possible sensors.
    *
    * NOTE: ARGoS creates and initializes actuators and sensors
    * internally, on the basis of the lists provided the configuration
    * file at the <controllers><eyebot_circle><actuators> and
    * <controllers><eyebot_circle><sensors> sections. If you forgot to
    * list a device in the XML and then you request it here, an error
    * occurs.
    */
   m_pcPosAct      = GetActuator<CCI_QuadRotorPositionActuator>("quadrotor_position");
   m_pcPosSens     = GetSensor  <CCI_PositioningSensor        >("positioning"       );
   m_pcRABSens     = GetSensor  <CCI_RangeAndBearingSensor    >("range_and_bearing" );
   m_pcLEDs        = GetActuator<CCI_LEDsActuator             >("leds"              );
   m_pcLightSensor = GetSensor	<CCI_EyeBotLightSensor              >("eyebot_light"           ); 
   /* 
    m_pcRotorSpeed = GetActuator<CCI_QuadRotorSpeedActuator>("quadrotor_speed"   ); 	
    */
   
  try {
	  
     GetNodeAttribute(t_node, "Xloc", XAxisCircle);
     GetNodeAttribute(t_node, "ALT", ALTITUDE);
     GetNodeAttribute(t_node, "yloc", YAxisCircle);
     GetNodeAttribute(t_node, "circleradius", CIRCLE_RADIUS);
     GetNodeAttribute(t_node, "circlewaypoints", CIRCLE_WAYPOINTS);
     GetNodeAttribute(t_node, "ywaypoints", Y_WAYPOINTS);
       
      GetNodeAttribute(t_node, "initial_rest_to_explore_prob", InitialRestToExploreProb);
      GetNodeAttribute(t_node, "initial_explore_to_rest_prob", InitialExploreToRestProb);
      GetNodeAttribute(t_node, "food_rule_explore_to_rest_delta_prob", FoodRuleExploreToRestDeltaProb);
      GetNodeAttribute(t_node, "food_rule_rest_to_explore_delta_prob", FoodRuleRestToExploreDeltaProb);
      GetNodeAttribute(t_node, "collision_rule_explore_to_rest_delta_prob", CollisionRuleExploreToRestDeltaProb);
      GetNodeAttribute(t_node, "social_rule_rest_to_explore_delta_prob", SocialRuleRestToExploreDeltaProb);
      GetNodeAttribute(t_node, "social_rule_explore_to_rest_delta_prob", SocialRuleExploreToRestDeltaProb);
      GetNodeAttribute(t_node, "minimum_resting_time", MinimumRestingTime);
      GetNodeAttribute(t_node, "minimum_unsuccessful_explore_time", MinimumUnsuccessfulExploreTime);
      GetNodeAttribute(t_node, "minimum_search_for_place_in_nest_time", MinimumSearchForPlaceInNestTime);
      
      GetNodeAttribute(t_node, "energy_per_robot", m_nEnergy);
     
     
   }
    /* Get the XLOC for the robot */
       catch(CARGoSException& ex) {
      THROW_ARGOSEXCEPTION_NESTED("Error initializing controller diffusion parameters.", ex);
   }
   Reset();
}

/****************************************/
/****************************************/

void CEyeBotCircle::ControlStep() {
   /* Get RAB message, if any */
   RLOG << "Message received: ";
   if(! m_pcRABSens->GetReadings().empty()) {
      m_psFBMsg = &(m_pcRABSens->GetReadings()[0]);
      LOG << *reinterpret_cast<const UInt64*>(m_psFBMsg->Data.ToCArray());
   }
   else {
      m_psFBMsg = NULL;
      LOG << "none_quadcopter";
   }
   LOG << std::endl;
   /* Execute state logic */
   switch(m_eState) {
      case STATE_START:
         TakeOff();
         break;
      case STATE_TAKE_OFF:
         TakeOff();
         break;
      case STATE_START_SERACH_POS:
         StartSearchPostion();
         break;
      case STATE_MOVE_ALONG_CIRCLE:
         MoveAlongCircle();
         break;
      case STATE_GO_TO_CENTER:
         GoToCenter();
         break;
      case STATE_LAND:
         Land();
         break;
      case STATE_RECHARGE:
	 Recharge();
	 break;
      default:
         LOGERR << "[BUG] Unknown robot state: " << m_eState << std::endl;
   }
   /* Write debug information */
   RLOG << "Current state: " << m_eState << std::endl;

  /* RLOG << "m_cTargetPos: " << m_cTargetPos << std::endl;
   RLOG << "m_cTargetPosFinal: " << m_cTargetPosFinal  << std::endl;
   RLOG << "m_circle: " << m_cCircleCenter  << std::endl;*/
  
  RLOG << "recharstate: " << m_sFoodData.rechargestates << std::endl;
   
   RLOG << "SFoodData: " << m_sFoodData.nEnergy << std::endl;
   RLOG << "m_energy: " << m_nEnergy - m_sFoodData.nEnergy << std::endl;
      
  CQuadcopterLoopFunctions* trailpheu = new CQuadcopterLoopFunctions;
   CQuadcopterLoopFunctions::Trailphe& trailphe = trailpheu->GetTrailData();  
   
   /*RLOG << "trailphe.LocationX: " << trailphe[5].LocationX << std::endl;*/
   RLOG << "trailphe.phe: " << trailphe.phe << std::endl;
  
   
   RLOG  << std::endl;
   
  /* RLOG << "m_zlightsensor" <<m_zlightsensor << std::endl;*/
   
   
}

/****************************************/
/****************************************/

void CEyeBotCircle::Reset() {
   /* Start the behavior */
   m_eState = STATE_START;
   /* No message received */
   m_psFBMsg = NULL;
   
   m_XWAYPOINTSLICE = m_NOXWAYPOINT_SLICE;
   m_pcLEDs->SetAllColors(CColor::RED);
   
   InNest = true;
   RestToExploreProb = InitialRestToExploreProb;
   ExploreToRestProb = InitialExploreToRestProb;
   TimeExploringUnsuccessfully = 0;
   /* Initially the robot is resting, and by setting RestingTime to
      MinimumRestingTime we force the robots to make a decision at the
      experiment start. If instead we set RestingTime to zero, we would
      have to wait till RestingTime reaches MinimumRestingTime before
      something happens, which is just a waste of time. */
   TimeRested = MinimumRestingTime;
   TimeSearchingForPlaceInNest = 0;
    /*
     m_pcRotorSpeed->SetLinearVelocity(CVector3(2, 2,2));
     */
    

  

}

/****************************************/
/****************************************/

void CEyeBotCircle::TakeOff() {
  
   if(m_eState != STATE_TAKE_OFF) {
      /* State initialization */
      m_eState = STATE_TAKE_OFF;
      m_cCircleCenter = m_pcPosSens->GetReading().Position + CVector3(0.0f, 0.0f, ALTITUDE);
      m_cTargetPos = m_cCircleCenter;
      m_pcPosAct->SetAbsolutePosition(m_cTargetPos);
      /*const CCI_LightSensor::std::vector<Real>& tLightReads = m_pcLightSensor->GetReadings();
      const CCI_FootBotLightSensor::TReadings& tLightReads = m_pcLight->GetReadings();*/
	/*m_zlightsensor = m_pcLightSensor->GetReadings();*/
   }
   else {
      if(Distance(m_cTargetPos, m_pcPosSens->GetReading().Position) < PROXIMITY_TOLERANCE) {
         /* State transition */
	  m_pcLEDs->SetAllColors(CColor::GREEN);
         StartSearchPostion();
      }
   }
}

/****************************************/
/****************************************/
 
void CEyeBotCircle::StartSearchPostion() {
   if(m_eState != STATE_START_SERACH_POS) {	

      /* State initialization */
      m_eState = STATE_START_SERACH_POS;
      m_funWaypoint = 0.1f;
      m_fYunWaypoint = 0.1f;
      m_searchlineno = 0;
      m_pcLEDs->SetAllColors(CColor::BLUE);
      
      m_pcRNGtarget = CRandom::CreateRNG("argos");
      m_rdmlocx = m_pcRNGtarget->Uniform(m_workingstartx);
      m_rdmlocy = m_pcRNGtarget->Uniform(m_workingstartx);
      m_cTargetPosFinal = m_pcPosSens->GetReading().Position + CVector3(m_rdmlocx, m_rdmlocy, 0.0f);
      m_rdmlocx = m_cTargetPosFinal[0];
      m_rdmlocy = m_cTargetPosFinal[1];
   }
   else {
     
	if ((m_nEnergy/2 - m_sFoodData.nEnergy) < 2 ){GoToCenter();}
	 

	  
	 m_cTargetPos = m_pcPosSens->GetReading().Position + CVector3(m_funWaypoint, m_fYunWaypoint, 0.0f);
         m_pcPosAct->SetAbsolutePosition(m_cTargetPos);
	 
	 m_sensx = m_pcPosSens->GetReading().Position[0];
	 m_sensy = m_pcPosSens->GetReading().Position[1];
	 
	 
	 if((m_rdmlocx - m_sensx) < 1.0f and (m_rdmlocy - m_sensy) < 1.0f) {
	   MoveAlongCircle();}
	   
	 else if ((m_rdmlocx - m_sensx) < 1.0f){
	   m_funWaypoint = 0.0f;
	    if ((m_rdmlocy - m_sensy) < 1.0f) {
	    m_fYunWaypoint = 0.0f;
	      MoveAlongCircle();
	    }
	}
	 else if ((m_rdmlocy - m_sensy) < 1.0f) {
	 m_fYunWaypoint = 0.0f;
	    if((m_rdmlocx - m_sensx) < 1.0f){
	    m_funWaypoint = 0.0f;
	    MoveAlongCircle();
	    }
	}
	 else{}

     /*
      m_cTargetPos = m_pcPosSens->GetReading().Position + CVector3(m_pcRNGtarget->Uniform(m_workingstartx),
								   m_pcRNGtarget->Uniform(m_workingstartx), 0.0f);         
     */
      /*m_eState = STATE_START_SERACH_POS;
      /*m_pcPosAct->SetAbsolutePosition(m_cTargetPos);*/
      
     
     /* if(Distance(m_cTargetPosFinal, m_pcPosSens->GetReading().Position) < PROXIMITY_TOLERANCE) {
         /* State transition 
	
	  
         MoveAlongCircle();
	 
      }*/
   }
}
/****************************************/
/****************************************/

/****************************************/
/****************************************/
/* Delete HERE FOR UNMASK !!!!!!!!!!!!
 
void CEyeBotCircle::LeaveCircleCenter() {
   if(m_eState != STATE_LEAVE_CIRCLE_CENTER) {
      /* State initialization */
      
/* Delete HERE FOR UNMASK !!!!!!!!!!!!

      m_eState = STATE_LEAVE_CIRCLE_CENTER;
      m_cTargetPos = m_pcPosSens->GetReading().Position + CVector3(CIRCLE_RADIUS, 0.0f, 0.0f);
      m_pcPosAct->SetAbsolutePosition(m_cTargetPos);
   }
   else {
      if(Distance(m_cTargetPos, m_pcPosSens->GetReading().Position) < PROXIMITY_TOLERANCE) {
         /* State transition */
	 
/* Delete HERE FOR UNMASK !!!!!!!!!!!!
	  
         MoveAlongCircle();
	 
      }
   }
}

/****************************************/
/****************************************/

void CEyeBotCircle::MoveAlongCircle() {
   if(m_eState != STATE_MOVE_ALONG_CIRCLE) {
      /* State initialization */
      m_eState = STATE_MOVE_ALONG_CIRCLE;
      m_unWaypoint = 0;
      m_YunWaypoint = 0;
        m_searchlineno = 0;
   }
   else {
     
     if ((m_nEnergy/2 - m_sFoodData.nEnergy) < 2){GoToCenter();} 
     
      if(m_searchlineno == 5) {
         /* State transition */
         GoToCenter();
      }
      else {
         /* State logic */
	 if (CIRCLE_WAYPOINTS>=m_unWaypoint){
         m_cTargetPos = m_pcPosSens->GetReading().Position + CVector3(m_XWAYPOINTSLICE, 0.0f, 0.0f);
         m_pcPosAct->SetAbsolutePosition(m_cTargetPos);
         ++m_unWaypoint;
	 }
	else {

         m_cTargetPos = m_pcPosSens->GetReading().Position + CVector3(0.0f, Y_WAYPOINTSLICE, 0.0f);
         m_pcPosAct->SetAbsolutePosition(m_cTargetPos);
	 ++m_YunWaypoint;
	if (Y_WAYPOINTS==m_YunWaypoint){ 
	  ++m_searchlineno;
	  m_unWaypoint = 0;
	  m_YunWaypoint = 0;	
	  m_XWAYPOINTSLICE *= -1.0f;
	}
	else{}
		 
	  }
	}
 }    
}
/****************************************/
/****************************************/

void CEyeBotCircle::GoToCenter() {
   m_pcLEDs->SetAllColors(CColor::YELLOW);
   if(m_eState != STATE_GO_TO_CENTER) {
      /* State initialization */
      m_eState = STATE_GO_TO_CENTER;
      m_funWaypoint = -0.1f;
      m_fYunWaypoint = -0.1f;
      m_rdmlocx = m_cCircleCenter[0];
      m_rdmlocy = m_cCircleCenter[1];
      m_cTargetPosFinal = m_cCircleCenter;
   }
   else {

	
	m_cTargetPos = m_pcPosSens->GetReading().Position + CVector3(m_funWaypoint, m_fYunWaypoint, 0.0f);
         m_pcPosAct->SetAbsolutePosition(m_cTargetPos);
	 
	 m_sensx = m_pcPosSens->GetReading().Position[0];
	 m_sensy = m_pcPosSens->GetReading().Position[1];
	 
	 
	 if(-(m_rdmlocx - m_sensx) < 1.0f and -(m_rdmlocy - m_sensy) < 1.0f) {
	   Land();}
	   
	 else if (-(m_rdmlocx - m_sensx) < 1.0f){
	   m_funWaypoint = 0.0f;
	    if (-(m_rdmlocy - m_sensy) < 1.0f) {
	    m_fYunWaypoint = 0.0f;
	      Land();
	    }
	}
	 else if (-(m_rdmlocy - m_sensy) < 1.0f) {
	 m_fYunWaypoint = 0.0f;
	    if(-(m_rdmlocx - m_sensx) < 1.0f){
	    m_funWaypoint = 0.0f;
	    Land();
	    }
	}
	 else{}

         /* State transition */
         
      }
   }

/****************************************/
/****************************************/

void CEyeBotCircle::Land() {
   if(m_eState != STATE_LAND) {
      /* State initialization */
      m_eState = STATE_LAND;
      m_cTargetPos = m_pcPosSens->GetReading().Position;
      m_cTargetPos.SetZ(0.0f);
      m_pcPosAct->SetAbsolutePosition(m_cTargetPos);
      
   }
   if(Distance(m_cTargetPos, m_pcPosSens->GetReading().Position) < PROXIMITY_TOLERANCE) {
         /* State transition */
	 m_pcLEDs->SetAllColors(CColor::RED);
         Recharge();
      }
}

/****************************************/
/****************************************/

void CEyeBotCircle::Recharge() {
     if(m_eState != STATE_RECHARGE) {
      /* State initialization */
      m_eState = STATE_RECHARGE;
      m_pcLEDs->SetAllColors(CColor::RED);
   }
   else{
  if (m_sFoodData.rechargestates == true) {
    
     /* Start the behavior */
   m_eState = STATE_START;
   /* No message received */
   m_psFBMsg = NULL;
   
   m_XWAYPOINTSLICE = m_NOXWAYPOINT_SLICE;
   m_pcLEDs->SetAllColors(CColor::RED);
   
   InNest = true;
   RestToExploreProb = InitialRestToExploreProb;
   ExploreToRestProb = InitialExploreToRestProb;
   TimeExploringUnsuccessfully = 0;
   /* Initially the robot is resting, and by setting RestingTime to
      MinimumRestingTime we force the robots to make a decision at the
      experiment start. If instead we set RestingTime to zero, we would
      have to wait till RestingTime reaches MinimumRestingTime before
      something happens, which is just a waste of time. */
   TimeRested = MinimumRestingTime;
   TimeSearchingForPlaceInNest = 0;
    /*
     m_pcRotorSpeed->SetLinearVelocity(CVector3(2, 2,2));
     */
  }
   }
  
  
}

/*
 * This statement notifies ARGoS of the existence of the controller.
 * It binds the class passed as first argument to the string passed as
 * second argument.
 * The string is then usable in the configuration file to refer to this
 * controller.
 * When ARGoS reads that string in the configuration file, it knows which
 * controller class to instantiate.
 * See also the configuration files for an example of how this is used.
 */
REGISTER_CONTROLLER(CEyeBotCircle, "quadcopter_controller")
