#ifndef QUADCOPTER_H
#define QUADCOPTER_H

/*
 * Include some necessary headers.
 */
/* Definition of the CCI_Controller class. */
#include <argos3/core/control_interface/ci_controller.h>
/* Definition of the quadrotor positioning actuator */
#include <argos3/plugins/robots/generic/control_interface/ci_quadrotor_position_actuator.h>
/* Definition of the positioning sensor */
#include <argos3/plugins/robots/generic/control_interface/ci_positioning_sensor.h>
/* Definition of the range-and-bearing sensor */
#include <argos3/plugins/robots/generic/control_interface/ci_range_and_bearing_sensor.h>
/* Definition of the LEDs actuator */
#include <argos3/plugins/robots/generic/control_interface/ci_leds_actuator.h>
/* Definition of the quadcopter speed actuator */
#include <argos3/plugins/robots/generic/control_interface/ci_quadrotor_speed_actuator.h>
/* Definitions for random number generation */
#include <argos3/core/utility/math/rng.h>

#include <argos3/plugins/robots/eye-bot/control_interface/ci_eyebot_light_sensor.h>
#include <argos3/plugins/robots/generic/control_interface/ci_light_sensor.h>

/*
 * All the ARGoS stuff in the 'argos' namespace.
 * With this statement, you save typing argos:: every time.
 */
using namespace argos;

/*
 * A controller is simply an implementation of the CCI_Controller class.
 */
class CEyeBotCircle : public CCI_Controller {

public:

   /* Class constructor. */
   CEyeBotCircle();
      
   
   struct SFoodData {
      bool HasFoodItem;      // true when the robot is carrying a food item
      bool rechargestates;
      size_t FoodItemIdx;    // the index of the current food item in the array of available food items
      size_t TotalFoodItems; // the total number of food items carried by this robot during the experiment
      UInt32 nEnergy;

      SFoodData();
      void Reset();
   };

   /* Class destructor. */
   virtual ~CEyeBotCircle() {}

   /*
    * This function initializes the controller.
    * The 't_node' variable points to the <parameters> section in th	e XML
    * file in the <controllers><eyebot_circle_controller> section.
    */
   virtual void Init(TConfigurationNode& t_node);

   /*
    * This function is called once every time step.
    * The length of the time step is set in the XML file.
    */
   virtual void ControlStep();

   /*
    * This function resets the controller to its state right after the
    * Init().
    * It is called when you press the reset button in the GUI.
    */
   virtual void Reset();

   /*
    * Called to cleanup what done by Init() when the experiment finishes.
    * In this example controller there is no need for clean anything up,
    * so the function could have been omitted. It's here just for
    * completeness.
    */
   
   virtual void Destroy() {}
   
      /*
    * Returns true if the robot is currently resting.
    */
   inline bool IsResting() const {
     /* 
      return m_sStateData.State == SStateData::STATE_RESTING; 
      */
     return m_eState == STATE_RECHARGE;
   }
   inline SFoodData& GetFoodData() {
     
      return m_sFoodData;
   }

   
private:

   /*
    * Takes off the robot.
    */
   void TakeOff();

   /*
    * Moves the robot towards the start postion
   */ 
   
     void StartSearchPostion();

   /*
    * Moves the robot along the circle.
    */
   void MoveAlongCircle();

   /*
    * Moves the robot to the center.
    */
   void GoToCenter();

   /*
    * Lands the robot.
    */
   void Land();
   
    /*
    * Recharge the robot.
    */
   void Recharge();

private:

   /* Current robot state */
   enum EState {
      STATE_START = 0,
      STATE_TAKE_OFF,
      STATE_START_SERACH_POS,
      STATE_MOVE_ALONG_CIRCLE,
      STATE_GO_TO_CENTER,
      STATE_LAND,
      STATE_RECHARGE
   };

private:

        /* True when the robot is in the nest */
      bool InNest;

      /* Initial probability to switch from resting to exploring */
      Real InitialRestToExploreProb;
      /* Current probability to switch from resting to exploring */
      Real RestToExploreProb;
      /* Initial probability to switch from exploring to resting */
      Real InitialExploreToRestProb;
      /* Current probability to switch from exploring to resting */
      Real ExploreToRestProb;
      /* Used as a range for uniform number generation */
      CRange<Real> ProbRange;
      /* The increase of ExploreToRestProb due to the food rule */
      Real FoodRuleExploreToRestDeltaProb;
      /* The increase of RestToExploreProb due to the food rule */
      Real FoodRuleRestToExploreDeltaProb;
      /* The increase of ExploreToRestProb due to the collision rule */
      Real CollisionRuleExploreToRestDeltaProb;
      /* The increase of RestToExploreProb due to the social rule */
      Real SocialRuleRestToExploreDeltaProb;
      /* The increase of ExploreToRestProb due to the social rule */
      Real SocialRuleExploreToRestDeltaProb;
      /* The minimum number of steps in resting state before the robots
         starts thinking that it's time to move */
      size_t MinimumRestingTime;
      /* The number of steps in resting state */
      size_t TimeRested;
      /* The number of exploration steps without finding food after which
         a foot-bot starts thinking about going back to the nest */
      size_t MinimumUnsuccessfulExploreTime;
      /* The number of exploration steps without finding food */
      size_t TimeExploringUnsuccessfully;
      /* If the robots switched to resting as soon as it enters the nest,
         there would be overcrowding of robots in the border between the
         nest and the rest of the arena. To overcome this issue, the robot
         spends some time looking for a place in the nest before finally
         settling. The following variable contains the minimum time the
         robot must spend in state 'return to nest' looking for a place in
         the nest before switching to the resting state. */
      size_t MinimumSearchForPlaceInNestTime;
      /* The time spent searching for a place in the nest */
      size_t TimeSearchingForPlaceInNest;
  
  
   /* Pointer to the quadrotor position actuator */
   CCI_QuadRotorPositionActuator* m_pcPosAct;
   /* Pointer to the positioning sensor */
   CCI_PositioningSensor* m_pcPosSens;
   /* Pointer to the range-and-bearing sensor */
   CCI_RangeAndBearingSensor* m_pcRABSens;
    /* Pointer to the LEDs actuator */
   CCI_LEDsActuator* m_pcLEDs;   
   CRandom::CRNG* m_pcRNG;
   /* Pointer to the quadcopter speed actuator */
   CCI_QuadRotorSpeedActuator* m_pcRotorSpeed; 
   /* Pointer to the quadrotor light sensor (z-rot only) */
   CCI_EyeBotLightSensor   * m_pcLightSensor;

   /* Current robot state */
   EState m_eState;
   /* Circle center */
   CVector3 m_cCircleCenter;
   /* Current target position */
   CVector3 m_cTargetPosFinal;
   CVector3 m_cTargetPos;
   /* Used to move the robot along the circle */
   UInt32 m_unWaypoint;
      /* Used to move the robot along the Y axis */
   UInt32 m_YunWaypoint;
         /* Used to set which line the robot is */
   UInt32 m_searchlineno;
    /* To store the Energy level of the robot */
   UInt32 m_nEnergy;
   
  /* Light sesnor working variable */
  std::vector<Real> m_zlightsensor;
 
   
   Real m_NOXWAYPOINT_SLICE;
   Real m_XWAYPOINTSLICE;
   Real m_CIRCLE_WAYPOINTSX;
   Real m_CIRCLE_WAYPOINTSY;
   Real m_rdmlocx;
   Real m_rdmlocy;
   Real m_sensx;
   Real m_sensy;
   Real m_funWaypoint;
   Real m_fYunWaypoint;
   
   CRandom::CRNG* m_pcRNGtarget;
   CRange<Real> m_workingstartx;
   
   SFoodData m_sFoodData;



   /* Contains the message received from the foot-bot */
   const CCI_RangeAndBearingSensor::SPacket* m_psFBMsg;
};

#endif
